# encoding=utf-8
"""
mwtemplates
Copyright (c) 2012-2014 Dan Michael O. Heggø

Simple wikitext template parser and editor
"""

from mwtemplates.templateeditor2 import TemplateEditor
from mwtemplates.preprocessor import preprocessToXml
